/* gmp_p1.c
 *
 * Implementation of Williams' p+1 algorithm using the GNU MP library.
 */

#include "gmp_p1.h"

static int gmp_factor_p1_a(mpz_t p, mpz_t n, unsigned int a,
unsigned int max_m) {
	mpz_t b, x, y;
	unsigned int k, m;
	int retval = 1;

	mpz_init_set_ui(b, a);
	mpz_init(x);
	mpz_init(y);
	m = 1;

	while(max_m--) {
		mpz_set(x, b);
		mpz_mul(y, b, b);
		mpz_sub_ui(y, y, 2);
		mpz_mod(y, y, n);
		k = 1;
		while(k <= m) k = k << 1;
		k = k >> 2;
		while(k) {
			if(m & k) {
				mpz_mul(x, x, y);
				mpz_sub(x, x, b);
				mpz_mod(x, x, n);
				mpz_mul(y, y, y);
				mpz_sub_ui(y, y, 2);
				mpz_mod(y, y, n);
			} else {
				mpz_mul(y, y, x);
				mpz_sub(y, y, b);
				mpz_mod(y, y, n);
				mpz_mul(x, x, x);
				mpz_sub_ui(x, x, 2);
				mpz_mod(x, x, n);
			}
			k = k >> 1;
		}
		mpz_set(b, x);
		mpz_sub_ui(x, x, 2);
		mpz_gcd(p, x, n);
		if(mpz_cmp_ui(p, 1) && mpz_cmp(p, n)) {
			retval = 0;
			break;
		}
		m++;
	}

	mpz_clear(b);
	mpz_clear(x);
	mpz_clear(y);
	return retval;
}

/* Factor the number n into a (not necessarily prime) factor p.
 * At most max_a different sequences will be tried, and for each sequence at
 * most max_m elements will be tried.
 * Return 0 on success, or 1 on failure on all sequences.
 */
int gmp_factor_p1(mpz_t p, mpz_t n, unsigned int max_m, unsigned int max_a) {
	unsigned int a;

	for(a=3; a<max_a+3; a++) {
		if(!gmp_factor_p1_a(p, n, a, max_m)) return 0;
	}
	return 1;
}

